/*
 * yasapi.h
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#ifndef __YASAPI_H__ // {
#define __YASAPI_H__
#include <yasapi_guid.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <math.h>
#include <winamp/wa_ipc.h>
#include <winamp/out.h>
#ifdef __cpluplus
extern "C" {
#endif

///////////////////////////////////////////////////////////////////////////////
// win32 thread-safe queue implementation using native windows API
// http://stackoverflow.com/questions/11706985/win32-thread-safe-queue-implementation-using-native-windows-api

// Mutex vs. Semaphores – Part 1: Semaphores
// https://blog.feabhas.com/2009/09/mutex-vs-semaphores-%E2%80%93-part-1-semaphores/

// Mutex vs. Semaphores – Part 2: The Mutex
// https://blog.feabhas.com/2009/09/mutex-vs-semaphores-%E2%80%93-part-2-the-mutex/

// Mutex vs. Semaphores – Part 3 (final part): Mutual Exclusion Problems
// https://blog.feabhas.com/2009/10/mutex-vs-semaphores-%E2%80%93-part-3-final-part-mutual-exclusion-problems/

// Task Synchronisation
// https://blog.feabhas.com/2009/10/task-synchronisation/

// Task Synchronisation – Part 2: Multiple Tasks and RTOS APIs
// https://blog.feabhas.com/2009/11/task-synchronisation-part-2-multiple-tasks-and-rtos-apis/

///////////////////////////////////////////////////////////////////////////////
//#define YASAPI_CO_INITIALIZE
//#define YASAPI_THREAD_CO_INITIALIZE
#define YASAPI_ABOUT
#define YASAPI_GAPLESS
#define YASAPI_SORROUND
#define YASAPI_RING_REALLOC
#define YASAPI_RING_DWORD
#define YASAPI_RING_BUF_SIZE
//#define YASAPI_CHANNEL_LAYOUT
//#define YASAPI_NOTIFY

///////////////////////////////////////////////////////////////////////////////
#if defined (_MSC_VER) && ! defined (__func__) // {
#define __func__                __FUNCTION__
#endif // }

#define YASAPI_WIDEN2(x)        L ## x
#define YASAPI_WIDEN(x)         YASAPI_WIDEN2(x)

#if ! defined (__WFILE__) // {
#define __WFILE__               YASAPI_WIDEN(__FILE__)
#endif // }

#if ! defined (__wfunc__) // {
#define __wfunc__               YASAPI_WIDEN(__func__)
#endif // }

// http://stackoverflow.com/questions/2192416/how-to-convert-concatenated-strings-to-wide-char-with-the-c-preprocessor
#define YASAPI_STR1(x)          #x
#define YASAPI_STR(x)           YASAPI_STR1(x)

#define YASAPI_WSTR1(x)         L ## #x
#define YASAPI_WSTR(x)          YASAPI_WSTR1(x)

// http://stackoverflow.com/questions/7795514/evaluate-preprocessor-token-before-concatenation
#define YASAPI_CONCAT1(A,B)     A ## B
#define YASAPI_CONCAT(A,B)      YASAPI_CONCAT1(A,B)

///////////////////////////////////////////////////////////////////////////////
#if ! defined (YASAPI_VER) // {
  #define YASAPI_VER            0.0.0
#endif // }

#define YASAPI_VERSION          YASAPI_STR(YASAPI_VER)
#define PI_VER2                 "v" YASAPI_VERSION

#if defined (OUT_YASAPI_SFX) // {
  #define YASAPI_PREFIX         YASAPI_CONCAT(out_yasapi,OUT_YASAPI_SFX)
#else // } {
  #define YASAPI_PREFIX         out_yasapi
#endif // }

#define YASAPI_ID_BASE          65536
#define YASAPI_ID_OFFS          858

#if defined (YASAPI_DEBUG) && defined (YASAPI_SSE2) // {
  #define PI_VER                PI_VER2 " (Debug/SSE2)"
  #define YASAPI_ID             YASAPI_PREFIX ## -debug-sse2
  #define YASAPI_WA_ID          (YASAPI_ID_BASE+YASAPI_ID_OFFS+3)
#elif defined (YASAPI_DEBUG) // {
  #define PI_VER                PI_VER2 " (Debug)"
  #define YASAPI_ID             YASAPI_PREFIX ## -debug
  #define YASAPI_WA_ID          (YASAPI_ID_BASE+YASAPI_ID_OFFS+0)
#elif defined (YASAPI_SSE2) // } {
  #define PI_VER                PI_VER2 " (SSE2)"
  #define YASAPI_ID             YASAPI_PREFIX ## -sse2
  #define YASAPI_WA_ID          (YASAPI_ID_BASE+YASAPI_ID_OFFS+1)
#else // } {
  #define PI_VER                PI_VER2
  #define YASAPI_ID             YASAPI_PREFIX
  #define YASAPI_WA_ID          (YASAPI_ID_BASE+YASAPI_ID_OFFS+2)
#endif // }

#define YASAPI_NAME             YASAPI_STR(YASAPI_ID)
#define YASAPI_MODULE           YASAPI_NAME ".dll"
#define YASAPI_PROPERTY_FILE    L"Plugins\\" YASAPI_WSTR(YASAPI_ID) L".ini"
#define PI_LABEL                "Yet Another (WA)SAPI Output " PI_VER

#define YASAPI_INT_MAX(n)       (~(~0ull<<((n)-1)))

#define YASAPI_SHAREMODE_LABEL(eShareMode) \
    (AUDCLNT_SHAREMODE_SHARED==(eShareMode)?"SHARED":"EXCLUSIVE")

#if 0 // {
#if ! defined (YASAPI_AUDCLNT_SHAREMODE) // {
#define YASAPI_MODE_LABEL(mode) ( \
    YASAPI_MODE_SHARED==(mode)?"SHARED": \
    YASAPI_MODE_EXCLUSIVE==(mode)?"EXCLUSIVE": \
    "AUTOMATIC" \
)
#endif // }
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_DEBUG) // {
  #define GETDEBUG(debug)     (debug)
#else // } {
  #define GETDEBUG(debug)     0
#endif // }

  #define DUNKNOWN(debug,y) \
    message(GETDEBUG(debug),1,0,&y,"UNKNOWN: \"%s\" (%d).\n", \
        basename(__FILE__),__LINE__)
  #define DERROR(debug,x,y) \
    message(GETDEBUG(debug),0,x,&y,#x ": \"%s\" (%d).\n", \
        basename(__FILE__),__LINE__)
  #define DERRORV(debug,x,y,format,...) \
    message(GETDEBUG(debug),0,x,&y,#x ": \"%s\" (%d).\n", \
        basename(__FILE__),__LINE__,__VA_ARGS__)

#if defined (YASAPI_DEBUG) // {
  #define DPUTS(debug,level,cs) \
      (((level)<(debug))&&fputs(cs,stderr))
  #define DPRINTF(debug,level,cs,...) \
      (((level)<(debug))&&fprintf(stderr,cs,__VA_ARGS__))
  #define DPUTWS(debug,level,ws) \
      (((level)<(debug))&&fputws(ws,stderr))
  #define DWPRINTF(debug,level,ws,...) \
      (((level)<(debug))&&fwprintf(stderr,ws,__VA_ARGS__))
  #define DMESSAGE(debug,m) (debug&&fprintf(stderr,"Error " m ": " \
      "\"%s\" (%d).\n",basename(__FILE__),__LINE__))
  #define DMESSAGEV(debug,m,...) (debug&&fprintf(stderr,"Error " m \
      ": \"%s\" (%d).\n",__VA_ARGS__,basename(__FILE__),__LINE__))
  #define DWMESSAGEV(debug,m,...) ((debug)&&fwprintf(stderr,L"Error " m \
      L": \"%s\" (%d)\n",__VA_ARGS__,wbasename(__WFILE__),__LINE__))
#else // } {
  #define DPUTS(debug,level,cs)
  #define DPRINTF(debug,level,cs,...)
  #define DPUTWS(debug,level,ws)
  #define DWPRINTF(debug,level,ws,...)
  #define DMESSAGE(debug,m)
  #define DMESSAGEV(debug,m,...)
  #define DWMESSAGEV(debug,m,...)
#endif // }

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_FRAMES_MS(nFrames,pwfx) \
  MulDiv(1000,nFrames,(pwfx)->nSamplesPerSec)
#define YASAPI_FRAMES_TIMER(nFrames,pwfx) \
  (10000.0*1000*(nFrames)/(pwfx)->nSamplesPerSec)

#define YASAPI_MIN_BALANCE_COUNT  (8ull)

#define YASAPI_MIN_RING_SIZE      (1.0)
#define YASAPI_MAX_RING_SIZE      (10.0)

#define YASAPI_MIN_SHARE_SIZE     (1.0)
#define YASAPI_MAX_SHARE_SIZE     (20.0)

#if defined (YASAPI_UNDERFLOW) // {
#define YASAPI_MIN_UNDERFLOW      (0.0)
#define YASAPI_MAX_UNDERFLOW      (1.0)
#endif // }

///////////////////////////////////////////////////////////////////////////////
DEFINE_GUID(IID_IUnknown, 0x00000000, 0x0000, 0x0000, 0xc0,0x00,
    0x00,0x00,0x00,0x00,0x00,0x46);

DEFINE_GUID(IID_IAudioClient, 0x1cb9ad4c, 0xdbfa, 0x4c32, 0xb1,0x78,
    0xc2,0xf5,0x68,0xa7,0x03,0xb2);
DEFINE_GUID(IID_IAudioClock, 0xcd63314f, 0x3fba, 0x4a1b, 0x81,0x2c,
    0xef,0x96,0x35,0x87,0x28,0xe7);
DEFINE_GUID(IID_IAudioRenderClient, 0xf294acfc, 0x3146, 0x4483, 0xa7,0xbf,
    0xad,0xdc,0xa7,0xc2,0x60,0xe2);

DEFINE_GUID(CLSID_MMDeviceEnumerator, 0xbcde0395, 0xe52f, 0x467c, 0x8e,0x3d,
    0xc4,0x57,0x92,0x91,0x69,0x2e);
DEFINE_GUID(IID_IMMDeviceEnumerator, 0xa95664d2, 0x9614, 0x4f35, 0xa7,0x46,
    0xde,0x8d,0xb6,0x36,0x17,0xe6);
DEFINE_GUID(IID_IMMDevice, 0xd666063f, 0x1587, 0x4e43, 0x81,0xf1,
    0xb9,0x48,0xe8,0x07,0x36,0x3f);

DEFINE_GUID(IID_IMMNotificationClient, 0x7991eec9, 0x7e89, 0x4d85, 0x83,0x90,
    0x6c,0x70,0x3c,0xec,0x60,0xc0);
#if 0 // {
DEFINE_GUID(CLSID_MMPlayerNotificationClient, 0x1e45b183,0x7929,0x40e1,
    0xB1,0x9B,0x93,0x83,0x2A,0x44,0x49,0x24);
#endif // }

///////////////////////////////////////////////////////////////////////////////
__declspec( dllexport ) Out_Module * winampGetOutModule();
char *basename(char *s);
wchar_t *wbasename(wchar_t *s);
wchar_t *wapath(wchar_t *file);
int message(int debug, int force, HRESULT x, HRESULT *y, const char *m, ...);
int crash(const char *format, ...);
#if defined (YASAPI_DEBUG) // {
void ConsoleOpen(void);
void ConsoleClose(void);
#endif // }

void WFXXSetup(WAVEFORMATEXTENSIBLE *pwfxx, int samplerate, int numchannels,
    int bitspersamp, BOOL bSurround, BOOL bFloat);

///////////////////////////////////////////////////////////////////////////////
#define VALUE_INT(pData,offs) \
  (*(int *)(((char *)(pData))+(offs)))
#define VALUE_DOUBLE(pData,offs) \
  (*(double *)(((char *)(pData))+(offs)))
#define VALUE_STRING(pData,offs) \
  ((wchar_t *)(((char *)(pData))+(offs)))

#define PROPERTY_INT(pProperty,pData) \
  VALUE_INT(pData,(pProperty)->offs)
#define PROPERTY_DOUBLE(pProperty,pData) \
  VALUE_DOUBLE(pData,(pProperty)->offs)
#define PROPERTY_STRING(pProperty,pData) \
  VALUE_STRING(pData,(pProperty)->offs)

#define YASAPI_PROPERTY_SIZE    128
#define YASAPI_GROUP_COMMON     L"plugin"
#define YASAPI_PROPERTY_DEBUG   L"debug"
#define YASAPI_PROPERTY_VIS     L"visualization"
#define YASAPI_PROPERTY_PAGE    L"page"
#if defined (YASAPI_ABOUT) // {
#define YASAPI_PROPERTY_CONFIGX L"posx"
#define YASAPI_PROPERTY_CONFIGY L"posy"
#define YASAPI_PROPERTY_ABOUTX  L"aboutx"
#define YASAPI_PROPERTY_ABOUTY  L"abouty"
#else // } {
#define YASAPI_PROPERTY_POSX    L"posx"
#define YASAPI_PROPERTY_POSY    L"posy"
#endif // }

#if defined (YASAPI_CHANNEL_LAYOUT) // {
typedef enum _ChannelLayout ChannelLayout;
#endif // }
typedef enum _PlayerState PlayerState;

typedef struct _PropertyType PropertyType;
typedef struct _PropertyLoadConfig PropertyLoadConfig;
typedef struct _PropertySaveConfig PropertySaveConfig;
typedef struct _Property Property;

typedef struct _Result Result;
typedef struct _Store Store;
typedef struct _Request Request;
typedef struct _QueueStrategy QueueStrategy;
typedef struct _Queue Queue;
typedef struct _Ring Ring;

#if defined (YASAPI_GAPLESS) // {
typedef enum _OptionGaplessOffset OptionGaplessOffset;
#endif // }
typedef struct _OptionsCommon OptionsCommon;
typedef struct _OptionsDevice OptionsDevice;
typedef struct _Options Options;

typedef struct _ControlCheckBoxConfig ControlCheckBoxConfig;
typedef struct _ControlRadioButtonConfig ControlRadioButtonConfig;
typedef struct _ControlSliderConfig ControlSliderConfig;
typedef struct _ControlComboBoxList ControlComboBoxList;
typedef struct _ControlComboBoxConfig ControlComboBoxConfig;
typedef struct _ControlSliderCascadeList ControlSliderCascadeList;
typedef struct _ControlSliderCascadeConfig ControlSliderCascadeConfig;
typedef struct _ControlType ControlType;
typedef struct _Control Control;

typedef struct _PlayerDevice PlayerDevice;
typedef struct _Player Player;
typedef struct _Strategy Strategy;

typedef void (*player_on_event_t)(Player *pPlayer);
typedef int (*player_proc_t)(Player *pPlayer, Request *pRequest);

///////////////////////////////////////////////////////////////////////////////
#define VALUE_INT(pData,offs) \
  (*(int *)(((char *)(pData))+(offs)))
#define VALUE_DOUBLE(pData,offs) \
  (*(double *)(((char *)(pData))+(offs)))
#define VALUE_STRING(pData,offs) \
  ((wchar_t *)(((char *)(pData))+(offs)))

#define PROPERTY_INT(pProperty,pData) \
  VALUE_INT(pData,(pProperty)->offs)
#define PROPERTY_DOUBLE(pProperty,pData) \
  VALUE_DOUBLE(pData,(pProperty)->offs)
#define PROPERTY_STRING(pProperty,pData) \
  VALUE_STRING(pData,(pProperty)->offs)

#define YASAPI_PROPERTY_SIZE    128
#define YASAPI_GROUP_COMMON     L"plugin"
#define YASAPI_PROPERTY_DEBUG   L"debug"
#define YASAPI_PROPERTY_VIS     L"visualization"
#define YASAPI_PROPERTY_PAGE    L"page"
#define YASAPI_PROPERTY_POSX    L"posx"
#define YASAPI_PROPERTY_POSY    L"posy"

#if defined (YASAPI_CHANNEL_LAYOUT) // {
///////////////////////////////////////////////////////////////////////////////
enum _ChannelLayout {
  CHANNEL_LAYOUT_NULL,
  CHANNEL_LAYOUT_MONO,
  CHANNEL_LAYOUT_STEREO,
  CHANNEL_LAYOUT_3POINT1,             // CHANNEL_LAYOUT_QUAD
  CHANNEL_LAYOUT_3POINT1_SURROUND,    // CHANNEL_LAYOUT_SURROUND
  CHANNEL_LAYOUT_5POINT1,
  CHANNEL_LAYOUT_5POINT1_SORROUND,
  CHANNEL_LAYOUT_7POINT1,
  CHANNEL_LAYOUT_7POINT1_SURROUND,
  CHANNEL_LAYOUT_9POINT1,
  CHANNEL_LAYOUT_9POINT1_SURROUND
};
#endif // }

///////////////////////////////////////////////////////////////////////////////
struct _PropertyLoadConfig {
  int debug;
  const wchar_t *path;
  void *pData;
  const void *pDefault;
  const wchar_t *group;
  const wchar_t *pfx;
};

struct _PropertySaveConfig {
  int debug;
  const wchar_t *path;
  void *pData;
  const wchar_t *group;
};

struct _PropertyType {
  void (*Load)(const Property *pProperty, PropertyLoadConfig *c);
  void (*Save)(const Property *pProperty, const PropertySaveConfig *c);
};

struct _Property {
  wchar_t *key;
  const PropertyType *type;
  int offs;
  int size;
};

extern const PropertyType gcIntType;
extern const PropertyType gcDoubleType;
extern const PropertyType gcStringType;

void PropertySaveInt(const wchar_t *group, const wchar_t *key, int n,
    const wchar_t *path);

void PropertyLoad(const Property *pProperty, PropertyLoadConfig *c);
void PropertySave(const Property *pProperty, const PropertySaveConfig *c);

///////////////////////////////////////////////////////////////////////////////
struct _Result {
  Result *prPrev;
  HANDLE hEvent;
  int state;
};

int ResultCreate(Result *pResult, int debug);
void ResultDestroy(Result *pResult, int debug);

///////////////////////////////////////////////////////////////////////////////
enum {
  STORE_SIZE=12
};

struct _Store {
  Result *prTos;
  Result aResult[STORE_SIZE];
  HANDLE hAvailable;
  HANDLE hMutex;
#if defined (YASAPI_DEBUG) // {
  int nAvailable;
  int nMinAvailable;
#endif // }
};

int StoreCreate(Store *pStore, int debug);
void StoreDestroy(Store *pStore, int debug);

void StorePush(Store *pStore, Result *pResult);
Result *StorePop(Store *pStore);

void StoreLock(Store *pStore);
void StoreUnlock(Store *pStore);

Result *StoreGet(Store *pStore);
void StorePut(Store *pStore, Result *pResult);

///////////////////////////////////////////////////////////////////////////////
struct _Request {
  Result *pResult;
  volatile int exit;
  volatile player_proc_t PlayerProc;
  volatile va_list ap;
};

///////////////////////////////////////////////////////////////////////////////
struct _QueueStrategy {
  const char *name;
  HANDLE (*GetSemaphoreDown)(Queue *pQueue);
  HANDLE (*GetSemaphoreUp)(Queue *pQueue);
  BOOL (*IsAlertable)(Queue *pQueue);
  Request **(*GetRequest)(Queue *pQueue);
#if defined (YASAPI_DEBUG) // {
  void (*Inc)(Queue *pQueue);
  void (*Dec)(Queue *pQueue);
#endif // }
};

extern const QueueStrategy cqsWrite;
extern const QueueStrategy cqsRead;

///////////////////////////////////////////////////////////////////////////////
enum {
  QUEUE_SIZE=12
};

struct _Queue {
  Request aRequest[QUEUE_SIZE];
  Request *rp;
  Request *wp;
  Request *mp;
  HANDLE hAvailable;
  HANDLE hWritten;
  HANDLE hMutex;
#if defined (YASAPI_DEBUG) // {
  int nAvailable;
  int nMinAvailable;
#endif // }
};

int QueueCreate(Queue *pQueue, int debug);
void QueueDestroy(Queue *pQueue, int debug);

void QueueLockMutex(Queue *pQueue);
void QueueUnlockMutex(Queue *pQueue);

Request *QueueLock(Queue *pQueue, const QueueStrategy *pStrategy,
    HANDLE hEvent, void (*OnEvent)(void *data), void *data);
void QueueUnlock(Queue *pQueue, const QueueStrategy *pStrategy,
    Result *pResult);
Request *QueueLockWrite(Queue *pQueue, Result *pResult, int exit,
    player_proc_t PlayerProc, Player *pPlayer, va_list ap);
Request *QueueLockRead(Queue *pQueue, HANDLE hEvent,
    void (*OnEvent)(void *data), void *data);
void QueueUnlockRead(Queue *pQueue, Result *pResult);
void QueueWrite(Queue *pQueue, Store *pStore, int exit,
    player_proc_t PlayerProc, Player *pPlayer, va_list ap);

#if defined (YASAPI_DEBUG) // {
void QueueIncAvailable(Queue *pQueue);
void QueueDecAvailable(Queue *pQueue);
void QueueNopAvailable(Queue *pQueue);
#endif // }

///////////////////////////////////////////////////////////////////////////////
typedef void (*RingCopy)(PVOID pClient, PVOID Destination,
    const VOID *Source, SIZE_T Length);

enum {
  RING_COPY=1,
  RING_COMMIT=2
};

struct _Ring {
#if defined (YASAPI_DEBUG) // {
  int *pDebug;
#endif // }
  RingCopy Copy;
  PVOID pClient;
  LPSTR pRep;
  LPSTR pMax;
  LPSTR pHead;
  LPSTR pTail;
  SIZE_T dwWritten;
  SIZE_T dwAvailable;
};

DWORD RingCreate(Ring *pRing, SIZE_T dwSize, int *pDebug);
void RingDestroy(Ring *pRing);

void RingReset(Ring *pRing);
#if defined (YASAPI_RING_REALLOC) // {
int RingRealloc(Ring *pRing, SIZE_T uSize, int *pDebug);
int RingReallocAvailable(Ring *pRing, SIZE_T dwAvailable, SIZE_T n);
#endif // }
#if defined (YASAPI_RING_DWORD) // {
DWORD RingWrite(Ring *pRing, LPCSTR pData, SIZE_T dwSize, SIZE_T n);
DWORD RingReadEx(Ring *pRing, LPSTR pData, SIZE_T dwSize, UINT uFlags);
DWORD RingRead(Ring *pRing, LPSTR pData, SIZE_T dwSize);
#else // } {
SIZE_T RingWrite(Ring *pRing, LPCSTR pData, SIZE_T dwSize, SIZE_T n);
SIZE_T RingReadEx(Ring *pRing, LPSTR pData, SIZE_T dwSize, UINT uFlags);
SIZE_T RingRead(Ring *pRing, LPSTR pData, SIZE_T dwSize);
#endif // }

void RingCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length);

///////////////////////////////////////////////////////////////////////////////
enum _OptionsSize {
  YASAPI_ID_SIZE=128
};

struct _OptionsCommon {
  // Persistent fields have to be of type "int", "double" or "wchar_t[]".
#if defined (YASAPI_DEBUG) // {
  int nDebug;
#endif // }
  wchar_t szId[YASAPI_ID_SIZE];
  int bMono2Stereo;
  int bVolume;
#if defined (YASAPI_SORROUND) // {
  int bSorround;
#endif // }
#if defined (YASAPI_GAPLESS) // {
  int bGapless;
#else // } {
  int nWriteBlock;
#endif // }
  int bFormatSupported;
  int bAudioClock;
  int bVisualization;
  int nPage;
#if defined (YASAPI_ABOUT) // {
  int nConfigX;
  int nConfigY;
  int nAboutX;
  int nAboutY;
#else // } {
  int nPosX;
  int nPosY;
#endif // }
#if defined (YASAPI_DEBUG) // {
  int nSleep;
#endif // }
};

int OptionsCommonLoad(OptionsCommon *pOptions, const wchar_t *pfx,
    const wchar_t *path, int debug);
void OptionsCommonSave(OptionsCommon *pOptions, const wchar_t *path,
    int debug);
const OptionsCommon *OptionsCommonDefault(void);

///////////////////////////////////////////////////////////////////////////////
enum _OptionsSharemode {
  YASAPI_SHAREMODE_SHARE,
  YASAPI_SHAREMODE_EXCLUSIVE,
  YASAPI_SHAREMODE_AUTOMATIC
};

enum _OptionsDevicePeriod {
  YASAPI_DEVICE_PERIOD_DEFAULT,
  YASAPI_DEVICE_PERIOD_MINIMUM,
};

#if defined (YASAPI_GAPLESS) // {
enum _OptionGaplessOffset {
  YASAPI_GAPLESS_OFFSET_POSITION,
  YASAPI_GAPLESS_OFFSET_TIME
};
#endif // }

struct _OptionsDevice {
  // Persistent fields have to be of type "int", "double" or "wchar_t[]".
  int eShareMode;
  int eDevicePeriod;
  int bAutoConvertPCM;
  int bSRCDefaultQuality;
  int bPull;
  double qShareSize;
#if defined (YASAPI_GAPLESS) // {
  int nGaplessOffset;
#endif // }

  struct {
    double qSize;
    double qFill;
  } ring;
};

int OptionsDeviceLoad(OptionsDevice *pOptions, const wchar_t *pfx,
    const wchar_t *pstrId, const wchar_t *path, int debug);
void OptionsDeviceSave(OptionsDevice *pOptions, const wchar_t *pstrId,
    const wchar_t *path, int debug);
const OptionsDevice *OptionsDeviceDefault(void);

///////////////////////////////////////////////////////////////////////////////
struct _Options {
  const wchar_t *path;
  OptionsCommon common;
  OptionsDevice device;
};

///////////////////////////////////////////////////////////////////////////////
struct _ControlType {
  const char *magic;
  const char *name;
  void (*Init)(const Control *pControl, HWND hDlg, HWND hWnd);
  void (*Set)(const Control *pControl, HWND hDlg, const void *pData);
  void (*Get)(const Control *pControl, HWND hDlg, void *pData);
  void (*Sync)(const Control *pControl, HWND hDlg);
};

struct _ControlCheckBoxConfig {
  int idc;
  int offset;
  int off;
  int on;
  wchar_t *help;
};

struct _ControlRadioButtonConfig {
  int idc;
  int offset;
  int val;
  wchar_t *help;
};

struct _ControlSliderConfig {
  int idcSlider;
  int offset;
  double min;
  double max;
  LONG lMaxRange;
  int idcStatic;
  const wchar_t *format;
  wchar_t *help;
};

struct _ControlComboBoxList {
  const wchar_t *label;
  int val;
};

struct _ControlComboBoxConfig {
  int idc;
  int offset;
  const ControlComboBoxList *list;
  wchar_t *help;
};

struct _ControlSliderCascadeList {
  int idcSlider;
  int offset;
  int idcStatic;
  const wchar_t *format;
  wchar_t *help;
};

struct _ControlSliderCascadeConfig {
  double min;
  double max;
  LONG lMaxRange;
  const ControlSliderCascadeList *list;
};

struct _Control {
  const ControlType *type;
  const void *config;
};

extern const ControlType gcCheckBoxType;
extern const ControlType gcRadioButtonType;
extern const ControlType gcSliderType;
extern const ControlType gcComboBoxType;
extern const ControlType gcSliderCascadeType;

HWND ControlCreateToolTip(HWND hDlg, HINSTANCE hInstance);
void ControlAddToolTip(HWND hDlg, HWND hWnd, int idc, PWSTR pszText);

///////////////////////////////////////////////////////////////////////////////
extern const Control gcaCommonControlList[];
extern const Control gcaDeviceControlList[];
extern const Control gcaCoreDeviceControlList[];
extern const Control gcaBuffersControlList[];

HWND ControlInit(const Control *pControl, HWND hDlg, HINSTANCE hInstance);
void ControlSet(const Control *pControl, HWND hDlg, const void *pData);
void ControlGet(const Control *pControl, HWND hDlg, void *pData);
void ControlSync(const Control *pControl, HWND hDlg);

///////////////////////////////////////////////////////////////////////////////
struct _PlayerDevice {
  IMMDevice *pDevice;
  wchar_t szId[YASAPI_ID_SIZE];
};

int PlayerDeviceCreate(PlayerDevice *pPlayerDevice, Player *pPlayer);
int PlayerDeviceDestroy(PlayerDevice *pPlayerDevice, Player *pPlayer);

int PlayerDeviceGet(PlayerDevice *pPlayerDevice, Player *pPlayer);

int PlayerDeviceCreateV(Player *pPlayer, Request *pRequst);
int PlayerDeviceDestroyV(Player *pPlayer, Request *pRequst);

///////////////////////////////////////////////////////////////////////////////
enum _PlayerState {
  PLAYER_STATE_NULL,
  PLAYER_STATE_BASE,
  PLAYER_STATE_RUN,
  PLAYER_STATE_OPEN,
  PLAYER_STATE_PAUSE,
  PLAYER_STATE_UNDERFLOW,
  PLAYER_STATE_PLAY,
  // 1. client indicates draining.
  PLAYER_STATE_DRAIN_1,
  // 2. ring buffer is empty (drained).
  PLAYER_STATE_DRAIN_2,
  // 3. player awaits read request in order to switch to eot.
  PLAYER_STATE_DRAIN_3,
  // end of track.
  PLAYER_STATE_EOT,
#if defined (YASAPI_GAPLESS) // {
  PLAYER_STATE_RESET
#endif // }
};

struct _Player {
  Options options;
  volatile HWND hDlgConfig;
  volatile PlayerState state;
#if defined (YASAPI_GAPLESS) // {
  volatile PlayerState state_save;
#endif // }

  struct {
    HINSTANCE hModule;
    char *pszModuleName;
    const char *pszFileName;
    double qVolume;
    Queue queue;
    Store store;
    HANDLE hEvent;
    HANDLE hTimer;
    HANDLE hThread;
  } base;

  struct {
    IMMDeviceEnumerator *pEnumerator;
#if defined (YASAPI_NOTIFY) // {
    IMMNotificationClient *pNotify;
#endif // }
  } run;

  PlayerDevice device;

  struct {
    const Strategy *pStrategy;
    WAVEFORMATEXTENSIBLE wfxx;
    WORD wBytesPerSample;
#if defined (YASAPI_SORROUND) // {
    int bSorround;
#endif // }
    int nShift;
    AUDCLNT_SHAREMODE eShareMode;
    IAudioClient *pClient;
    IAudioRenderClient *pRender;
    IAudioClock *pClock;
    UINT32 uFrames;
    UINT32 uFramesMin;
    LONGLONG time;
    Ring ring;
#if defined (YASAPI_RING_REALLOC) // {
    DWORD dwBufSize;
#endif // }

#if defined (YASAPI_GAPLESS) // {
    struct {
      int nGaplessOffset;

      union {
        struct {
          UINT64 u64Frequency;
          UINT64 u64Position;
        };

        struct {
          double qTime;
        };
      };
    };
#endif // }
  } open;
};

int PlayerCreate(Player *pPlayer, HINSTANCE hModule, const wchar_t *path,
    int debug);
void PlayerDestroy(Player *pPlayer);

int PlayerRun(Player *pPlayer, Request *pRequest);
int PlayerStop(Player *pPlayer, Request *pRequest);
int PlayerOpen(Player *pPlayer, Request *pRequest);
#if defined (YASAPI_GAPLESS) // {
int PlayerReset(Player *pPlayer, Request *pRequest);
#endif // }
int PlayerClose(Player *pPlayer, Request *pRequest);
int PlayerCanWrite(Player *pPlayer, Request *pRequest);
int PlayerWrite(Player *pPlayer, Request *pRequest);
int PlayerFlush(Player *pPlayer, Request *pRequest);
int PlayerIsPlaying(Player *pPlayer, Request *pRequest);
int PlayerPause(Player *pPlayer, Request *pRequest);
int PlayerGetTime(Player *pPlayer, Request *pRequest);
#if defined (YASAPI_GAPLESS) // {
int PlayerSetState(Player *pPlayer, Request *pRequest);
#endif // }
int PlayerRead(Player *pPlayer, Request *pRequest);

int PlayerDispatchSync(Player *pPlayer, int exit, player_proc_t PlayerProc,
    ...);
void PlayerDispatchAsync(void *data);
VOID CALLBACK PlayerFireTimer(LPVOID lpArg, DWORD dwLow, DWORD dwHigh);
void PlayerSetTimer(Player *pPlayer, LONGLONG time);
UINT32 PlayerGetFramesPadding(Player *pPlayer);
DWORD WINAPI PlayerThread(LPVOID lpParameter);

HWND PlayerGetDlgConfig(Player *pPlayer);
void PlayerPostUpdate(Player *pPlayer, WORD wRing, WORD wShared);

void PlayerCopyMemory(PVOID p, PVOID Destination, const VOID *Source,
    SIZE_T Length);

#if defined (YASAPI_NOTIFY) // {
int PlayerAddNotify(Player *pPlayer);
void PlayerRemoveNotify(Player *pPlayer);
#endif // }

///////////////////////////////////////////////////////////////////////////////
struct _Strategy {
  HRESULT (*Initialize)(Player *pPlayer, DWORD dwStreamFlags,
      REFERENCE_TIME hnsDevicePeriod);
  LONGLONG (*Interval)(Player *pPlayer, LONGLONG time);
  void (*SetTimer)(Player *pPlayer, LONGLONG time, BOOL bFlush);
  BOOL (*NeedDevicePeriod)(Player *pPlayer);
  BOOL (*NeedPadding)(Player *pPlayer);
  HRESULT (*SetEvent)(Player *pPlayer);
  HANDLE (*GetEvent)(Player *pPlayer);
  UINT32 (*GetFramesPadding)(Player *pPlayer);
};

extern const Strategy gcStrategyPush;
extern const Strategy gcStrategyPull;

///////////////////////////////////////////////////////////////////////////////
enum {
  WM_CONFIG_UPDATE=WM_APP
};

int ConfigDialog(Player *pPlayer, HINSTANCE hInstance, HWND hWndParent);
#if defined (YASAPI_ABOUT) // {
void AboutDialog(Player *pPlayer, HINSTANCE hInstance, HWND hWndParent);
#else // } {
void AboutDialog(HWND hWndParent);
#endif // }

#ifdef __cpluplus
}
#endif
#endif // }
